@@ -13,10 +13,6 @@
 # limitations under the License.
 """TensorFlow Privacy library."""
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-
 import sys
 
 from tensorflow_privacy.version import __version__  # pylint: disable=g-bad-import-order
@@ -27,24 +23,20 @@ if hasattr(sys, 'skip_tf_privacy_import'):  # Useful for standalone scripts.
   pass
 else:
   # TensorFlow v1 imports
-  import tensorflow_privacy.v1
+  from tensorflow_privacy import v1
 
   # Analysis
   from tensorflow_privacy.privacy.analysis.compute_dp_sgd_privacy_lib import compute_dp_sgd_privacy
-  from tensorflow_privacy.privacy.analysis.privacy_ledger import GaussianSumQueryEntry
-  from tensorflow_privacy.privacy.analysis.privacy_ledger import PrivacyLedger
-  from tensorflow_privacy.privacy.analysis.privacy_ledger import QueryWithLedger
-  from tensorflow_privacy.privacy.analysis.privacy_ledger import SampleEntry
-  from tensorflow_privacy.privacy.analysis.rdp_accountant import compute_heterogenous_rdp
-  from tensorflow_privacy.privacy.analysis.rdp_accountant import compute_rdp
-  from tensorflow_privacy.privacy.analysis.rdp_accountant import compute_rdp_from_ledger
-  from tensorflow_privacy.privacy.analysis.rdp_accountant import get_privacy_spent
+  from tensorflow_privacy.privacy.analysis.tree_aggregation_accountant import compute_rdp_tree_restart
+  from tensorflow_privacy.privacy.analysis.tree_aggregation_accountant import compute_rdp_single_tree
+  from tensorflow_privacy.privacy.analysis.tree_aggregation_accountant import compute_zcdp_single_tree
 
   # DPQuery classes
   from tensorflow_privacy.privacy.dp_query.dp_query import DPQuery
   from tensorflow_privacy.privacy.dp_query.dp_query import SumAggregationDPQuery
   from tensorflow_privacy.privacy.dp_query.discrete_gaussian_query import DiscreteGaussianSumQuery
   from tensorflow_privacy.privacy.dp_query.distributed_discrete_gaussian_query import DistributedDiscreteGaussianSumQuery
+  from tensorflow_privacy.privacy.dp_query.distributed_skellam_query import DistributedSkellamSumQuery
   from tensorflow_privacy.privacy.dp_query.gaussian_query import GaussianSumQuery
   from tensorflow_privacy.privacy.dp_query.nested_query import NestedQuery
   from tensorflow_privacy.privacy.dp_query.no_privacy_query import NoPrivacyAverageQuery
@@ -53,9 +45,12 @@ else:
   from tensorflow_privacy.privacy.dp_query.quantile_estimator_query import QuantileEstimatorQuery
   from tensorflow_privacy.privacy.dp_query.quantile_estimator_query import NoPrivacyQuantileEstimatorQuery
   from tensorflow_privacy.privacy.dp_query.quantile_adaptive_clip_sum_query import QuantileAdaptiveClipSumQuery
+  from tensorflow_privacy.privacy.dp_query import restart_query
   from tensorflow_privacy.privacy.dp_query import tree_aggregation
+  from tensorflow_privacy.privacy.dp_query.restart_query import RestartQuery
   from tensorflow_privacy.privacy.dp_query.tree_aggregation_query import TreeCumulativeSumQuery
   from tensorflow_privacy.privacy.dp_query.tree_aggregation_query import TreeResidualSumQuery
+  from tensorflow_privacy.privacy.dp_query.tree_range_query import TreeRangeSumQuery
 
   # Estimators
   from tensorflow_privacy.privacy.estimators.dnn import DNNClassifier
@@ -76,12 +71,4 @@ else:
   from tensorflow_privacy.privacy.optimizers.dp_optimizer_keras_vectorized import VectorizedDPKerasSGDOptimizer
   from tensorflow_privacy.privacy.optimizers.dp_optimizer_keras_vectorized import make_vectorized_keras_optimizer_class
 
-  try:
-    from tensorflow_privacy.privacy.bolt_on.models import BoltOnModel
-    from tensorflow_privacy.privacy.bolt_on.optimizers import BoltOn
-    from tensorflow_privacy.privacy.bolt_on.losses import StrongConvexMixin
-    from tensorflow_privacy.privacy.bolt_on.losses import StrongConvexBinaryCrossentropy
-    from tensorflow_privacy.privacy.bolt_on.losses import StrongConvexHuber
-  except ImportError:
-    # module `bolt_on` not yet available in this version of TF Privacy
-    pass
+  # module `bolt_on` not yet available in this version of TF Privacy
