@@ -13,15 +13,8 @@
 # limitations under the License.
 """Differentially private optimizers for TensorFlow."""
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-
 from absl import logging
-
-import tensorflow.compat.v1 as tf
-
-from tensorflow_privacy.privacy.analysis import privacy_ledger
+import tensorflow as tf
 from tensorflow_privacy.privacy.dp_query import gaussian_query
 
 
@@ -35,12 +28,12 @@ def make_optimizer_class(cls):
   Returns:
     A DP-SGD subclass of `cls`.
   """
-  parent_code = tf.train.Optimizer.compute_gradients.__code__
+  parent_code = tf.compat.v1.train.Optimizer.compute_gradients.__code__
 
   has_compute_gradients = hasattr(cls, 'compute_gradients')
   if has_compute_gradients:
     child_code = cls.compute_gradients.__code__
-  GATE_OP = tf.train.Optimizer.GATE_OP  # pylint: disable=invalid-name
+  GATE_OP = tf.compat.v1.train.Optimizer.GATE_OP  # pylint: disable=invalid-name
   if has_compute_gradients and child_code is not parent_code:
     logging.warning(
         'WARNING: Calling make_optimizer_class() on class %s that overrides '
@@ -99,6 +92,7 @@ def make_optimizer_class(cls):
         dp_sum_query,
         num_microbatches=None,
         unroll_microbatches=False,
+        while_loop_parallel_iterations=10,
         *args,  # pylint: disable=keyword-arg-before-vararg, g-doc-args
         **kwargs):
       """Initializes the DPOptimizerClass.
@@ -112,10 +106,14 @@ def make_optimizer_class(cls):
         unroll_microbatches: If true, processes microbatches within a Python
           loop instead of a `tf.while_loop`. Can be used if using a
           `tf.while_loop` raises an exception.
+        while_loop_parallel_iterations: The number of iterations allowed to run
+          in parallel. It must be a positive integer. Applicable only when
+          unroll_microbatches is set to False. It gives users some control over
+          memory consumption.
         *args: These will be passed on to the base class `__init__` method.
         **kwargs: These will be passed on to the base class `__init__` method.
       """
-      super(DPOptimizerClass, self).__init__(*args, **kwargs)
+      super().__init__(*args, **kwargs)
       self._dp_sum_query = dp_sum_query
       self._num_microbatches = num_microbatches
       self._global_state = None
@@ -123,6 +121,7 @@ def make_optimizer_class(cls):
       # Beware: When num_microbatches is large (>100), enabling this parameter
       # may cause an OOM error.
       self._unroll_microbatches = unroll_microbatches
+      self._while_loop_parallel_iterations = while_loop_parallel_iterations
       self._was_compute_gradients_called = False
 
     def compute_gradients(self,
@@ -165,9 +164,9 @@ def make_optimizer_class(cls):
         for idx in range(self._num_microbatches):
           sample_state = process_microbatch(idx, sample_state)
 
-        grad_sums, self._global_state = (
-            self._dp_sum_query.get_noised_result(
-                sample_state, self._global_state))
+        grad_sums, self._global_state, _ = (
+            self._dp_sum_query.get_noised_result(sample_state,
+                                                 self._global_state))
 
         def normalize(v):
           return v / tf.cast(self._num_microbatches, tf.float32)
@@ -178,10 +177,6 @@ def make_optimizer_class(cls):
         return grads_and_vars
 
       else:
-        # TF is running in graph mode. Check we did not receive a gradient tape.
-        if gradient_tape:
-          raise ValueError('When in graph mode, a tape should not be passed.')
-
         # Note: it would be closer to the correct i.i.d. sampling of records if
         # we sampled each microbatch from the appropriate binomial distribution,
         # although that still wouldn't be quite correct because it would be
@@ -197,8 +192,8 @@ def make_optimizer_class(cls):
           """Process one microbatch (record) with privacy helper."""
           self_super = super(DPOptimizerClass, self)
 
-          mean_loss = tf.reduce_mean(input_tensor=tf.gather(
-              microbatches_losses, [i]))
+          mean_loss = tf.reduce_mean(
+              input_tensor=tf.gather(microbatches_losses, [i]))
 
           if hasattr(self_super, 'compute_gradients'):
             # This case covers optimizers in tf.train.
@@ -207,10 +202,15 @@ def make_optimizer_class(cls):
             # This case covers Keras optimizers from optimizers_v2.
             compute_gradients_fn = self_super._compute_gradients  # pylint: disable=protected-access
 
-          grads, _ = zip(*compute_gradients_fn(
-              mean_loss, var_list, gate_gradients,
-              aggregation_method, colocate_gradients_with_ops, grad_loss))
-          grads_list = list(grads)
+          if gradient_tape:
+            # This is intended to work for TF2 and may not work for TF1.
+            with gradient_tape.stop_recording():
+              grads_list = list(gradient_tape.gradient(mean_loss, var_list))
+          else:
+            grads, _ = zip(*compute_gradients_fn(
+                mean_loss, var_list, gate_gradients, aggregation_method,
+                colocate_gradients_with_ops, grad_loss))
+            grads_list = list(grads)
 
           sample_state = self._dp_sum_query.accumulate_record(
               sample_params, sample_state, grads_list)
@@ -218,8 +218,8 @@ def make_optimizer_class(cls):
 
         if var_list is None:
           var_list = (
-              tf.trainable_variables() + tf.get_collection(
-                  tf.GraphKeys.TRAINABLE_RESOURCE_VARIABLES))
+              tf.compat.v1.trainable_variables() + tf.compat.v1.get_collection(
+                  tf.compat.v1.GraphKeys.TRAINABLE_RESOURCE_VARIABLES))
 
         sample_state = self._dp_sum_query.initial_sample_state(var_list)
 
@@ -234,11 +234,14 @@ def make_optimizer_class(cls):
           body_fn = lambda i, state: [tf.add(i, 1), process_microbatch(i, state)]  # pylint: disable=line-too-long
           idx = tf.constant(0)
           _, sample_state = tf.while_loop(
-              cond=cond_fn, body=body_fn, loop_vars=[idx, sample_state])
+              cond=cond_fn,
+              body=body_fn,
+              loop_vars=[idx, sample_state],
+              parallel_iterations=self._while_loop_parallel_iterations)
 
-        grad_sums, self._global_state = (
-            self._dp_sum_query.get_noised_result(
-                sample_state, self._global_state))
+        grad_sums, self._global_state, _ = (
+            self._dp_sum_query.get_noised_result(sample_state,
+                                                 self._global_state))
 
         def normalize(v):
           try:
@@ -258,8 +261,8 @@ def make_optimizer_class(cls):
           ' called. Which means that the training is not differentially '
           'private. It happens for example in Keras training in TensorFlow '
           '2.0+.')
-      return super(DPOptimizerClass,
-                   self).apply_gradients(grads_and_vars, global_step, name)
+      return super(DPOptimizerClass, self).apply_gradients(
+          grads_and_vars=grads_and_vars, global_step=global_step, name=name)
 
   return DPOptimizerClass
 
@@ -278,7 +281,7 @@ def make_gaussian_optimizer_class(cls):
   class DPGaussianOptimizerClass(make_optimizer_class(cls)):  # pylint: disable=empty-docstring
     __doc__ = ("""DP subclass of `{}`.
 
-       You can use this as a differentially private replacement for 
+       You can use this as a differentially private replacement for
        `tf.compat.v1.train.{}`. This optimizer implements DP-SGD using
        the standard Gaussian mechanism.
 
@@ -290,7 +293,7 @@ def make_gaussian_optimizer_class(cls):
 
        ```python
        # Create optimizer.
-       opt = {}(l2_norm_clip=1.0, noise_multiplier=0.5, num_microbatches=1, 
+       opt = {}(l2_norm_clip=1.0, noise_multiplier=0.5, num_microbatches=1,
                 <standard arguments>)
        ```
 
@@ -307,9 +310,7 @@ def make_gaussian_optimizer_class(cls):
        ```
 
        """).format(
-           'tf.compat.v1.train.' + cls.__name__,
-           cls.__name__,
-           cls.__name__,
+           'tf.compat.v1.train.' + cls.__name__, cls.__name__, cls.__name__,
            'DP' + cls.__name__.replace('Optimizer', 'GaussianOptimizer'))
 
     def __init__(
@@ -317,7 +318,6 @@ def make_gaussian_optimizer_class(cls):
         l2_norm_clip,
         noise_multiplier,
         num_microbatches=None,
-        ledger=None,
         unroll_microbatches=False,
         *args,  # pylint: disable=keyword-arg-before-vararg
         **kwargs):
@@ -329,7 +329,6 @@ def make_gaussian_optimizer_class(cls):
         num_microbatches: Number of microbatches into which each minibatch is
           split. If `None`, will default to the size of the minibatch, and
           per-example gradients will be computed.
-        ledger: Defaults to `None`. An instance of `tf_privacy.PrivacyLedger`.
         unroll_microbatches: If true, processes microbatches within a Python
           loop instead of a `tf.while_loop`. Can be used if using a
           `tf.while_loop` raises an exception.
@@ -344,16 +343,9 @@ def make_gaussian_optimizer_class(cls):
       dp_sum_query = gaussian_query.GaussianSumQuery(
           l2_norm_clip, l2_norm_clip * noise_multiplier)
 
-      if ledger:
-        dp_sum_query = privacy_ledger.QueryWithLedger(dp_sum_query,
-                                                      ledger=ledger)
-
-      super(DPGaussianOptimizerClass, self).__init__(
-          dp_sum_query,
-          num_microbatches,
-          unroll_microbatches,
-          *args,
-          **kwargs)
+      super(DPGaussianOptimizerClass,
+            self).__init__(dp_sum_query, num_microbatches, unroll_microbatches,
+                           *args, **kwargs)
 
     def get_config(self):
       """Creates configuration for Keras serialization.
@@ -370,20 +362,18 @@ def make_gaussian_optimizer_class(cls):
       config.update({
           'l2_norm_clip': self._l2_norm_clip,
           'noise_multiplier': self._noise_multiplier,
-          'num_microbatches': self._num_microbatches})
+          'num_microbatches': self._num_microbatches
+      })
 
       return config
 
-    @property
-    def ledger(self):
-      return self._dp_sum_query.ledger
-
   return DPGaussianOptimizerClass
 
-AdagradOptimizer = tf.train.AdagradOptimizer
-AdamOptimizer = tf.train.AdamOptimizer
-GradientDescentOptimizer = tf.train.GradientDescentOptimizer
-RMSPropOptimizer = tf.train.RMSPropOptimizer
+
+AdagradOptimizer = tf.compat.v1.train.AdagradOptimizer
+AdamOptimizer = tf.compat.v1.train.AdamOptimizer
+GradientDescentOptimizer = tf.compat.v1.train.GradientDescentOptimizer
+RMSPropOptimizer = tf.compat.v1.train.RMSPropOptimizer
 
 DPAdagradOptimizer = make_optimizer_class(AdagradOptimizer)
 DPAdamOptimizer = make_optimizer_class(AdamOptimizer)
