@@ -14,10 +14,6 @@
 # ==============================================================================
 """Differentially private version of Keras optimizer v2."""
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-
 import tensorflow as tf
 
 from tensorflow_privacy.privacy.dp_query import gaussian_query
@@ -49,7 +45,7 @@ def make_keras_optimizer_class(cls):
 
     ```python
     # Create optimizer.
-    opt = {dp_keras_class}(l2_norm_clip=1.0, noise_multiplier=0.5, num_microbatches=1, 
+    opt = {dp_keras_class}(l2_norm_clip=1.0, noise_multiplier=0.5, num_microbatches=1,
             <standard arguments>)
     ```
 
@@ -81,9 +77,43 @@ def make_keras_optimizer_class(cls):
     model.fit(...)
     ```
 
-    """.format(base_class='tf.keras.optimizers.' + cls.__name__,
-               short_base_class=cls.__name__,
-               dp_keras_class='DPKeras' + cls.__name__)
+    In DP-SGD training, a larger batch size typically helps to achieve better
+    privacy/utility tradeoff. However there is typically a maximum batch size
+    imposed by hardware.
+    This optimizer can emulate large batch sizes on hardware with limited
+    memory by accumulating gradients for several steps before actually
+    applying them to update model weights.
+    Constructor argument `gradient_accumulation_steps` controls the number
+    of steps for which gradients are accumulated before updating
+    the model weights.
+
+    Below is an example which demonstrates how to use this feature:
+
+    ```python
+    # Create optimizer which will be accumulating gradients for 4 steps.
+    # and then performing an update of model weights.
+    opt = {dp_keras_class}(l2_norm_clip=1.0,
+                           noise_multiplier=0.5,
+                           num_microbatches=1,
+                           gradient_accumulation_steps=4,
+                           <standard arguments>)
+
+    # Use optimizer in a regular way.
+    # First three calls to opt.minimize won't update model weights and will
+    # only accumulate gradients. Model weights will be updated on the fourth
+    # call to opt.minimize
+    opt.minimize(loss, var_list=[var])
+    ```
+
+    Note that when using this feature effective batch size is
+    `gradient_accumulation_steps * one_step_batch_size` where
+    `one_step_batch_size` size of the batch which is passed to single step
+    of the optimizer. Thus user may have to adjust learning rate, weight decay
+    and possibly other training hyperparameters accordingly.
+    """.format(
+        base_class='tf.keras.optimizers.' + cls.__name__,
+        short_base_class=cls.__name__,
+        dp_keras_class='DPKeras' + cls.__name__)
 
     # The class tf.keras.optimizers.Optimizer has two methods to compute
     # gradients, `_compute_gradients` and `get_gradients`. The first works
@@ -99,6 +129,7 @@ def make_keras_optimizer_class(cls):
         l2_norm_clip,
         noise_multiplier,
         num_microbatches=None,
+        gradient_accumulation_steps=1,
         *args,  # pylint: disable=keyword-arg-before-vararg, g-doc-args
         **kwargs):
       """Initialize the DPOptimizerClass.
@@ -106,12 +137,22 @@ def make_keras_optimizer_class(cls):
       Args:
         l2_norm_clip: Clipping norm (max L2 norm of per microbatch gradients).
         noise_multiplier: Ratio of the standard deviation to the clipping norm.
-        num_microbatches: Number of microbatches into which each minibatch
-          is split.
+        num_microbatches: Number of microbatches into which each minibatch is
+          split. Default is `None` which means that number of microbatches
+          is equal to batch size (i.e. each microbatch contains exactly one
+          example). If `gradient_accumulation_steps` is greater than 1 and
+          `num_microbatches` is not `None` then the effective number of
+          microbatches is equal to
+          `num_microbatches * gradient_accumulation_steps`.
+        gradient_accumulation_steps: If greater than 1 then optimizer will be
+          accumulating gradients for this number of optimizer steps before
+          applying them to update model weights. If this argument is set to 1
+          then updates will be applied on each optimizer step.
         *args: These will be passed on to the base class `__init__` method.
         **kwargs: These will be passed on to the base class `__init__` method.
       """
-      super(DPOptimizerClass, self).__init__(*args, **kwargs)
+      super().__init__(*args, **kwargs)
+      self.gradient_accumulation_steps = gradient_accumulation_steps
       self._l2_norm_clip = l2_norm_clip
       self._noise_multiplier = noise_multiplier
       self._num_microbatches = num_microbatches
@@ -120,6 +161,64 @@ def make_keras_optimizer_class(cls):
       self._global_state = None
       self._was_dp_gradients_called = False
 
+    def _create_slots(self, var_list):
+      super()._create_slots(var_list)  # pytype: disable=attribute-error
+      if self.gradient_accumulation_steps > 1:
+        for var in var_list:
+          self.add_slot(var, 'grad_acc')
+
+    def _prepare_local(self, var_device, var_dtype, apply_state):
+      super()._prepare_local(var_device, var_dtype, apply_state)  # pytype: disable=attribute-error
+      if self.gradient_accumulation_steps > 1:
+        apply_update = tf.math.equal(
+            tf.math.floormod(self.iterations + 1,
+                             self.gradient_accumulation_steps), 0)
+        grad_scaler = tf.cast(1. / self.gradient_accumulation_steps, var_dtype)
+        apply_state[(var_device, var_dtype)].update({
+            'apply_update': apply_update,
+            'grad_scaler': grad_scaler
+        })
+
+    def _resource_apply_dense(self, grad, var, apply_state=None):
+      if self.gradient_accumulation_steps > 1:
+        var_device, var_dtype = var.device, var.dtype.base_dtype
+        coefficients = ((apply_state or {}).get((var_device, var_dtype)) or
+                        self._fallback_apply_state(var_device, var_dtype))
+        grad_acc = self.get_slot(var, 'grad_acc')
+
+        def _update_grad():
+          apply_grad_op = super(DPOptimizerClass, self)._resource_apply_dense(
+              grad_acc + grad * coefficients['grad_scaler'], var, apply_state)  # pytype: disable=attribute-error
+          with tf.control_dependencies([apply_grad_op]):
+            return grad_acc.assign(
+                tf.zeros_like(grad_acc),
+                use_locking=self._use_locking,
+                read_value=False)
+
+        def _accumulate():
+          return grad_acc.assign_add(
+              grad * coefficients['grad_scaler'],
+              use_locking=self._use_locking,
+              read_value=False)
+
+        return tf.cond(coefficients['apply_update'], _update_grad, _accumulate)
+      else:
+        return super()._resource_apply_dense(grad, var, apply_state)  # pytype: disable=attribute-error
+
+    def _resource_apply_sparse_duplicate_indices(self, *args, **kwargs):
+      if self.gradient_accumulation_steps > 1:
+        raise NotImplementedError(
+            'Sparse gradients are not supported with large batch emulation.')
+      else:
+        return super()._resource_apply_sparse_duplicate_indices(*args, **kwargs)  # pytype: disable=attribute-error
+
+    def _resource_apply_sparse(self, *args, **kwargs):
+      if self.gradient_accumulation_steps > 1:
+        raise NotImplementedError(
+            'Sparse gradients are not supported with large batch emulation.')
+      else:
+        return super()._resource_apply_sparse(*args, **kwargs)  # pytype: disable=attribute-error
+
     def _compute_gradients(self, loss, var_list, grad_loss=None, tape=None):
       """DP-SGD version of base class method."""
 
@@ -157,7 +256,8 @@ def make_keras_optimizer_class(cls):
 
       # Compute the per-microbatch losses using helpful jacobian method.
       with tf.keras.backend.name_scope(self._name + '/gradients'):
-        jacobian = tape.jacobian(microbatch_losses, var_list)
+        jacobian = tape.jacobian(
+            microbatch_losses, var_list, unconnected_gradients='zero')
 
         # Clip gradients to given l2_norm_clip.
         def clip_gradients(g):
@@ -210,7 +310,7 @@ def make_keras_optimizer_class(cls):
       sample_state = self._dp_sum_query.initial_sample_state(params)
       for idx in range(self._num_microbatches):
         sample_state = process_microbatch(idx, sample_state)
-      grad_sums, self._global_state = (
+      grad_sums, self._global_state, _ = (
           self._dp_sum_query.get_noised_result(sample_state,
                                                self._global_state))
 
@@ -224,7 +324,26 @@ def make_keras_optimizer_class(cls):
 
       return final_grads
 
-    def apply_gradients(self, grads_and_vars, global_step=None, name=None):
+    def get_config(self):
+      """Returns the config of the optimizer.
+
+      An optimizer config is a Python dictionary (serializable)
+      containing the configuration of an optimizer.
+      The same optimizer can be reinstantiated later
+      (without any saved state) from this configuration.
+
+      Returns:
+          Python dictionary.
+      """
+      config = super().get_config()
+      config.update({
+          'l2_norm_clip': self._l2_norm_clip,
+          'noise_multiplier': self._noise_multiplier,
+          'num_microbatches': self._num_microbatches,
+      })
+      return config
+
+    def apply_gradients(self, *args, **kwargs):
       """DP-SGD version of base class method."""
       assert self._was_dp_gradients_called, (
           'Neither _compute_gradients() or get_gradients() on the '
@@ -232,8 +351,7 @@ def make_keras_optimizer_class(cls):
           'training is not differentially private. It may be the case that '
           'you need to upgrade to TF 2.4 or higher to use this particular '
           'optimizer.')
-      return super(DPOptimizerClass,
-                   self).apply_gradients(grads_and_vars, global_step, name)
+      return super().apply_gradients(*args, **kwargs)
 
   return DPOptimizerClass
 
